import collections
import json
import pathlib
import typing
import warnings

import pandas
import pydantic
import pynwb
import typing_extensions

from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel
from ..notifications import Notification


def _infer_scalar_field(
    electrode_name_to_series: dict[str, list[pynwb.icephys.PatchClampSeries]], field_name: str
) -> dict[str, float | None]:
    """For icephys specifically, infer some scalar field (e.g., rate, gain) for each electrode name."""
    electrode_name_to_field_values = collections.defaultdict(set)
    for electrode_name, series_list in electrode_name_to_series.items():
        for series in series_list:
            field_value = getattr(series, field_name, None)
            if field_value is not None:
                electrode_name_to_field_values[electrode_name].add(field_value)

    electrode_name_to_field = dict()
    for electrode_name, field_values in electrode_name_to_field_values.items():
        if len(ls := list(field_values)) == 1:
            electrode_name_to_field[electrode_name] = ls[0]
        else:
            message = (
                f"Some PatchClampSeries associated with electrode {electrode_name} have conflicting "
                f"{field_name}s: {field_values}. Automatic detection of channel {field_name} for this "
                "case is not yet implemented."
            )
            warnings.warn(message=message, stacklevel=2)

    return electrode_name_to_field


class Channel(BaseMetadataModel):
    channel_name: str
    reference: str
    type: str = "N/A"
    unit: str = "V"
    sampling_frequency: float | None = None
    channel_label: str | None = None
    stream_id: str | None = None
    description: str | None = None
    status: typing.Literal["good", "bad"] | None = None
    status_description: str | None = None
    gain: float | None = None
    time_offset: float | None = None
    time_reference_channels: str | None = None
    ground: str | None = None
    # recording_mode: str | None = None  # TODO: icephys only


class ChannelTable(BaseMetadataContainerModel):
    channels: list[Channel]
    modality: typing.Literal["ecephys", "icephys"]

    @pydantic.computed_field
    @property
    def notifications(self) -> list[Notification]:
        """
        All notifications from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        notifications = [notification for channel in self.channels for notification in channel.notifications]
        notifications.sort(
            key=lambda notification: (-notification.category.value, -notification.severity.value, notification.title)
        )
        return notifications

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        has_ecephys_electrodes = nwbfile.electrodes is not None
        has_icephys_electrodes = any(nwbfile.icephys_electrodes)
        if not has_ecephys_electrodes and not has_icephys_electrodes:
            return None

        if has_ecephys_electrodes and has_icephys_electrodes:
            message = (
                "Converting electrode metadata when there are both ecephys and icephys types "
                "has not yet been implemented."
            )
            raise NotImplementedError(message)

        modality = "ecephys" if has_ecephys_electrodes else "icephys"
        if modality == "ecephys":
            # Only scan electrical series listed under acquisition since those under processing can downsample the rate
            sampling_frequency = None
            stream_id = None
            gain = None
            raw_electrical_series = [
                neurodata_object
                for neurodata_object in nwbfile.objects.values()
                if isinstance(neurodata_object, pynwb.ecephys.ElectricalSeries)
            ]
            if len(raw_electrical_series) > 1:
                # TODO: form a map of electrode to rate/gate based on ElectricalSeries linkage
                message = (
                    "Support for automatic extraction of rates/gains from multiple ElectricalSeries is not yet "
                    "implemented. Skipping `sampling_frequency`, `stream_id`, and `gain` extraction."
                )
                warnings.warn(message=message, stacklevel=2)

            if len(raw_electrical_series) == 1:
                electrical_series = raw_electrical_series[0]
                if electrical_series.rate is None:
                    message = (
                        "Support for automatic extraction of rate from ElectricalSeries with timestamps "
                        "is not yet implemented. Skipping `sampling_frequency`, `stream_id`, and `gain` extraction."
                    )
                    warnings.warn(message=message, stacklevel=2)

                sampling_frequency = electrical_series.rate
                stream_id = electrical_series.name
                gain = electrical_series.conversion

            channels = [
                Channel(
                    channel_name=(
                        f"ch{channel_name.values[0]}"
                        if (channel_name := electrode.get("channel_name", None)) is not None
                        else f"ch{electrode.index[0]}"
                    ),
                    reference=(
                        f"contact{contact_ids.values[0]}"  # TODO: do a deep dive into edge cases of this reference
                        if (contact_ids := electrode.get("contact_ids", None)) is not None
                        else f"e{electrode.index[0]}"
                    ),
                    type="N/A",  # TODO: in dedicated follow-up, could classify LFP based on container
                    unit="V",
                    sampling_frequency=sampling_frequency,
                    # channel_label: str | None = None # TODO: only support with additional metadata
                    stream_id=stream_id,
                    # description: str | None = None  # TODO: only support with additional metadata
                    # status: typing.Literal["good", "bad"] | None = None # TODO: only support with additional metadata
                    # status_description: str | None = None # TODO: only support with additional metadata
                    gain=gain,
                    # Special extraction from SpikeInterface field
                    time_offset=shift[0] if (shift := electrode.get("inter_sample_shift", None)) is not None else None,
                    # time_reference_channels: str | None = None # TODO: only support with additional metadata
                    # ground: str | None = None # TODO: only support with additional metadata
                )
                for electrode in nwbfile.electrodes
            ]
        else:
            icephys_series = [
                neurodata_object
                for neurodata_object in nwbfile.acquisition.values()
                if isinstance(neurodata_object, pynwb.icephys.PatchClampSeries)
            ]
            # TODO: handle intracellular_recordings case
            electrode_name_to_series = collections.defaultdict(list)
            for series in icephys_series:
                electrode_name_to_series[series.electrode.name].append(series)

            electrode_name_to_sampling_frequency = _infer_scalar_field(
                electrode_name_to_series=electrode_name_to_series, field_name="rate"
            )
            electrode_name_to_gain = _infer_scalar_field(
                electrode_name_to_series=electrode_name_to_series, field_name="gain"
            )
            electrode_name_to_class = _infer_scalar_field(
                electrode_name_to_series=electrode_name_to_series, field_name="__class__"
            )

            class_name_to_type = {
                "VoltageClampSeries": "VM",
                "CurrentClampSeries": "IM",
            }
            electrode_name_to_type = {
                electrode_name: class_name_to_type.get(series_class.__name__, "n/a")
                for electrode_name, series_class in electrode_name_to_class.items()
            }
            type_to_recording_mode = {
                "VM": "voltage-clamp",
                "IM": "current-clamp",
                "n/a": "n/a",
            }

            electrode_name_to_stream_ids = {
                electrode_name: ",".join([series.name for series in series_list])
                for electrode_name, series_list in electrode_name_to_series.items()
            }

            channels = [
                Channel(
                    channel_name=electrode.name,
                    reference="n/a",  # TODO: think about if/how this could be any other value
                    type=electrode_name_to_type.get(electrode.name, "n/a"),
                    unit="V",
                    sampling_frequency=electrode_name_to_sampling_frequency.get(electrode.name, None),
                    # channel_label: str | None = None # TODO: only support with additional metadata
                    stream_id=electrode_name_to_stream_ids.get(electrode.name, None),
                    # description: str | None = None  # TODO: only support with additional metadata
                    # status: typing.Literal["good", "bad"] | None = None # TODO: only support with additional metadata
                    # status_description: str | None = None # TODO: only support with additional metadata
                    gain=electrode_name_to_gain.get(electrode.name, None),
                    # time_offset=
                    # time_reference_channels: str | None = None # TODO: only support with additional metadata
                    # ground: str | None = None # TODO: only support with additional metadata
                    recording_mode=type_to_recording_mode[electrode_name_to_type.get(electrode.name, "n/a")],
                    # TODO: add extra columns
                )
                for electrode in nwbfile.icephys_electrodes.values()
            ]
        return cls(channels=channels, modality=modality)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path):
        """
        Write the channels data to a TSV file.

        Parameters
        ----------
        file_path : path
            The path where the TSV file will be saved.
        """
        data = []
        for channel in self.channels:
            model_dump = channel.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)
        data_frame = data_frame.dropna(axis=1, how="all")
        data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the channels information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(
                obj=dict(),  # TODO
                fp=file_stream,
                indent=4,
            )



import collections
import os
import pathlib
import shutil
import typing
import warnings

import pydantic
import typing_extensions

from ._datalad_utils import _content_is_retrieved
from ._run_config import RunConfig
from .._converters._base_converter import BaseConverter
from .._tools import cache_read_nwb
from ..bids_models import BidsSessionMetadata
from ..notifications import Notification


class SessionConverter(BaseConverter):
    """
    Initialize a converter of NWB files to BIDS format.
    """

    session_id: str = pydantic.Field(
        description="A unique session identifier.",
    )
    nwbfile_paths: list[pydantic.FilePath] | list[pydantic.HttpUrl] = pydantic.Field(
        description="List of file paths or URLs of NWB files which share this session ID.", min_length=1
    )
    session_metadata: BidsSessionMetadata | None = pydantic.Field(
        description="BIDS metadata extracted for this session.", default=None
    )
    notifications: list[Notification] = pydantic.Field(
        description="List of auto-detected suggestions.", ge=0, default_factory=list
    )
    modality: typing.Literal["ecephys", "icephys"] | None = pydantic.Field(
        description="The modality for this session - auto-detected during metadata extraction step.", default=None
    )

    @classmethod
    @pydantic.validate_call
    def from_nwb_paths(
        cls,
        nwb_paths: list[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
        run_config: RunConfig = pydantic.Field(default_factory=lambda: RunConfig()),
        ignore_hidden: bool = True,
    ) -> list[typing_extensions.Self]:
        """
        Initialize a list of session converters from a list of NWB file paths.

        Automatically associates multiple NWB files to a single session according to session ID.

        Parameters
        ----------
        nwb_paths : iterable of file and directory paths
            An iterable of NWB file paths and directories containing NWB files.
        run_config : RunConfig, optional
            The configuration for this conversion run.
        ignore_hidden : bool, default: True
            Whether to ignore NWB files located under directories (those starting with a period).

        Returns
        -------
        A list of SessionConverter instances, one per unique session ID.
        """
        all_nwbfile_paths = []
        for nwb_path in nwb_paths:
            if nwb_path.is_file():
                all_nwbfile_paths.append(nwb_path)
            elif nwb_path.is_dir():
                all_nwbfile_paths += [
                    path
                    for path in nwb_path.rglob(pattern="*.nwb")
                    # Ignore DataLad files not retrieved from the annex
                    if ".git" not in path.parts and _content_is_retrieved(file_path=path)
                ]

        nwbfile_paths_to_convert = all_nwbfile_paths
        if ignore_hidden:
            # Ignore contents in hidden folders
            # Such as .git since contains .git/annex which might include NWB extensions, DS_Store, etc.
            nwbfile_paths_to_convert = [
                path for path in all_nwbfile_paths if not any(part.startswith(".") for part in path.parts)
            ]

        unique_session_id_to_nwbfile_paths = collections.defaultdict(list)
        for nwbfile_path in nwbfile_paths_to_convert:
            unique_session_id_to_nwbfile_paths[cache_read_nwb(nwbfile_path).session_id].append(nwbfile_path)

        session_converters = [
            cls(
                session_id=session_id or "0",  # Always include `ses-` entity, even for single-session subjects
                nwbfile_paths=nwbfile_paths,
                run_config=run_config,
            )
            for session_id, nwbfile_paths in unique_session_id_to_nwbfile_paths.items()
        ]
        return session_converters

    def extract_metadata(self) -> None:
        if self.session_metadata is not None:
            return

        self.run_config.bids_directory.mkdir(exist_ok=True)
        self.run_config._nwb2bids_directory.mkdir(exist_ok=True)

        self.session_metadata = BidsSessionMetadata.from_nwbfile_paths(
            nwbfile_paths=self.nwbfile_paths, run_config=self.run_config
        )
        self.notifications += self.session_metadata.notifications

        if self.modality is not None:
            return None

        detected_modalities = set()
        for source in ["probe_table", "electrode_table", "channel_table"]:
            table = getattr(self.session_metadata, source)
            if table is not None:
                detected_modalities.add(table.modality)
        if len(detected_modalities) > 1:
            message = (
                "Conflicting modalities found among probe, electrode, and channel tables. "
                "Conversion when multiple modalities are present has not yet been implemented."
            )
            raise NotImplementedError(message)
        if len(detected_modalities) == 0:
            # TODO: may change to 'beh' in future
            message = "No modality information found in session metadata - assuming ecephys."
            # TODO: change to global notification
            warnings.warn(message=message, stacklevel=2)
            self.modality = "ecephys"
        else:
            self.modality = next(iter(detected_modalities))

    def convert_to_bids_session(self) -> None:
        """
        Convert the NWB file to a BIDS session directory.
        """
        if len(self.nwbfile_paths) > 1:
            message = (
                "Conversion of multiple NWB files per session is not yet supported. "
                "Please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss the use case."
            )
            raise NotImplementedError(message)

        if self.session_metadata is None:
            self.extract_metadata()

        participant_id = self.session_metadata.sanitization.sanitized_participant_id
        session_id = self.session_metadata.sanitization.sanitized_session_id
        file_prefix = f"sub-{participant_id}_ses-{session_id}"

        self.write_ephys_files()
        if self.session_metadata.events is not None:
            self.write_events_files()

        modality_directory = self._establish_modality_subdirectory()
        for nwbfile_path in self.nwbfile_paths:
            session_file_path = modality_directory / f"{file_prefix}_{self.modality}.nwb"

            # If not a local path, then it is a URL, so write simple 'symlink' pointing to the URL
            if not isinstance(nwbfile_path, pathlib.Path):
                with session_file_path.open(mode="w") as file_stream:
                    file_stream.write(str(nwbfile_path))
                continue

            if self.run_config.file_mode == "copy":
                shutil.copy(src=nwbfile_path, dst=session_file_path)
            elif self.run_config.file_mode == "move":
                shutil.move(src=nwbfile_path, dst=session_file_path)
            elif self.run_config.file_mode == "symlink":
                relative_target = os.path.relpath(nwbfile_path.resolve(), session_file_path.parent.resolve())
                session_file_path.symlink_to(target=relative_target)

    def write_ephys_files(self) -> None:
        """
        Write the `_probes`, `_channels`, and `_electrodes` metadata files, both `.tsv` and `.json`, for this session.
        """
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)

        if (
            self.session_metadata.probe_table is None
            and self.session_metadata.channel_table is None
            and self.session_metadata.electrode_table is None
        ):
            return

        participant_id = self.session_metadata.sanitization.sanitized_participant_id
        session_id = self.session_metadata.sanitization.sanitized_session_id
        file_prefix = f"sub-{participant_id}_ses-{session_id}"

        modality_directory = self._establish_modality_subdirectory()
        if self.session_metadata.probe_table is not None:
            probes_tsv_file_path = modality_directory / f"{file_prefix}_probes.tsv"
            self.session_metadata.probe_table.to_tsv(file_path=probes_tsv_file_path)

            probes_json_file_path = modality_directory / f"{file_prefix}_probes.json"
            self.session_metadata.probe_table.to_json(file_path=probes_json_file_path)

        if self.session_metadata.channel_table is not None:
            channels_tsv_file_path = modality_directory / f"{file_prefix}_channels.tsv"
            self.session_metadata.channel_table.to_tsv(file_path=channels_tsv_file_path)

            channels_json_file_path = modality_directory / f"{file_prefix}_channels.json"
            self.session_metadata.channel_table.to_json(file_path=channels_json_file_path)

        if self.session_metadata.electrode_table is not None:
            electrodes_tsv_file_path = modality_directory / f"{file_prefix}_electrodes.tsv"
            self.session_metadata.electrode_table.to_tsv(file_path=electrodes_tsv_file_path)

            electrodes_json_file_path = modality_directory / f"{file_prefix}_electrodes.json"
            self.session_metadata.electrode_table.to_json(file_path=electrodes_json_file_path)

    def write_events_files(self) -> None:
        """Write the `_events.tsv` and `_events.json` files for this session."""
        if self.session_metadata.events is None:
            message = "No events metadata found in the session metadata - unable to write events TSV."
            raise ValueError(message)
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)

        participant_id = self.session_metadata.sanitization.sanitized_participant_id
        session_id = self.session_metadata.sanitization.sanitized_session_id
        file_prefix = f"sub-{participant_id}_ses-{session_id}"

        ecephys_directory = self._establish_modality_subdirectory()
        session_events_tsv_file_path = ecephys_directory / f"{file_prefix}_events.tsv"
        self.session_metadata.events.to_tsv(file_path=session_events_tsv_file_path)

        # TODO: add merging from top-level additional metadata mechanism
        session_events_metadata_file_path = ecephys_directory / f"{file_prefix}_events.json"
        self.session_metadata.events.to_json(file_path=session_events_metadata_file_path)

    def _establish_modality_subdirectory(self) -> pathlib.Path:
        if self.modality is None:
            message = "Modality has not been determined for this session - unable to establish modality subdirectory."
            raise ValueError(message)

        participant_id = self.session_metadata.sanitization.sanitized_participant_id
        session_id = self.session_metadata.sanitization.sanitized_session_id

        subject_directory = self.run_config.bids_directory / f"sub-{participant_id}"
        subject_directory.mkdir(exist_ok=True)
        session_directory = subject_directory / f"ses-{session_id}"
        session_directory.mkdir(exist_ok=True)
        modality_directory = session_directory / self.modality
        modality_directory.mkdir(exist_ok=True)

        return modality_directory



import pathlib

import py

import nwb2bids


def test_icephys_tutorial_file(tmpdir: py.path.local, temporary_bids_directory: pathlib.Path):
    tutorial_path = nwb2bids.testing.generate_ephys_tutorial(
        output_directory=pathlib.Path(tmpdir), mode="file", modality="icephys"
    )

    nwb_paths = [tutorial_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-001"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-001": {
            "directories": {"ses-A"},
            "files": {"sub-001_sessions.json", "sub-001_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-001"
        / "ses-A": {
            "directories": {"icephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-001"
        / "ses-A"
        / "icephys": {
            "directories": set(),
            "files": {
                "sub-001_ses-A_icephys.nwb",
                "sub-001_ses-A_probes.json",
                "sub-001_ses-A_probes.tsv",
                "sub-001_ses-A_electrodes.json",
                "sub-001_ses-A_electrodes.tsv",
                "sub-001_ses-A_channels.json",
                "sub-001_ses-A_channels.tsv",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import os
import pathlib
import typing

import pydantic

from .._converters._dataset_converter import DatasetConverter
from .._converters._run_config import RunConfig


def _make_absolute(p: pathlib.Path) -> pathlib.Path:
    """Make path absolute and normalized without following symlinks.

    Uses os.path.abspath for normalization because:
    - pathlib.resolve() follows symlinks (breaks git-annex/datalad paths)
    - pathlib.absolute() doesn't normalize '..' components
    - os.path.abspath normalizes without following symlinks
    """
    return pathlib.Path(os.path.abspath(p))


AbsoluteFilePath = typing.Annotated[pydantic.FilePath, pydantic.BeforeValidator(_make_absolute)]
AbsoluteDirectoryPath = typing.Annotated[pydantic.DirectoryPath, pydantic.BeforeValidator(_make_absolute)]


@pydantic.validate_call
def convert_nwb_dataset(
    *,
    nwb_paths: list[AbsoluteFilePath | AbsoluteDirectoryPath] = pydantic.Field(min_length=1),
    run_config: RunConfig = pydantic.Field(default_factory=lambda: RunConfig()),
) -> DatasetConverter:
    """
    Convert a dataset of NWB files to a BIDS dataset.

    Parameters
    ----------
    nwb_paths : any iterable of file or directory paths
        An iterable of NWB file paths and directories containing NWB files.
    run_config : RunConfig, optional
        The configuration for this conversion run.

    Returns
    -------
    dataset_converter : DatasetConverter
        The DatasetConverter used to perform the conversion.
        Contains notifications and other contextual information about the conversion process.
    """
    dataset_converter = DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths, run_config=run_config)
    dataset_converter.extract_metadata()
    dataset_converter.convert_to_bids_dataset()

    return dataset_converter



"""Tests for message handling passed to the top-level `converter_nwb_dataset` function."""

import json
import pathlib

import nwb2bids


def test_notifications_baseline(minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [minimal_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)


def test_notifications_1(problematic_nwbfile_path_1: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_1]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    notifications = dataset_converter.notifications
    expected_notifications = [
        nwb2bids.Notification(
            title="Invalid species",
            reason="Participant species is not a proper Latin binomial or NCBI Taxonomy id.",
            solution=(
                "Rename the subject species to a Latin binomial, obolibrary taxonomy link, or NCBI taxonomy "
                "reference."
            ),
            examples=[],
            field="nwbfile.subject.species",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.SCHEMA_INVALIDATION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.Notification(
            title="Invalid participant ID",
            reason=(
                "The participant ID contains invalid characters. BIDS allows only the plus sign to be used as a "
                "separator in the subject entity label. Underscores, dashes, spaces, slashes, and other special "
                "characters (including #) are expressly forbidden."
            ),
            solution="Rename the subject without using any special characters except for `+`.",
            examples=["`ab_01` -> `ab+01`", "`subject #2` -> `subject+2`", "`id 2 from 9/1/25` -> `id+2+9+1+25`"],
            field="nwbfile.subject.subject_id",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.Notification(
            title="Invalid participant sex (BIDS)",
            reason="Participant sex is not one of the allowed patterns by BIDS.",
            solution="Rename the subject sex to be one of the accepted values.",
            examples=["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
            field="nwbfile.subject.sex",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.BIDS],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.Notification(
            title="Invalid participant sex (archives)",
            reason="Participant sex is not one of the allowed patterns by the common archives.",
            solution="Rename the subject sex to be one of the accepted values.",
            examples=["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
            field="nwbfile.subject.sex",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
    ]
    assert notifications == expected_notifications

    assert dataset_converter.run_config.notifications_json_file_path.exists()
    with dataset_converter.run_config.notifications_json_file_path.open(mode="r") as file_stream:
        notifications_json = json.load(fp=file_stream)
    str_nwb_paths = [str(path) for path in nwb_paths]
    expected_notification_json = [
        {
            "category": "SCHEMA_INVALIDATION",
            "data_standards": ["DANDI"],
            "examples": [],
            "field": "nwbfile.subject.species",
            "reason": "Participant species is not a proper Latin binomial or NCBI " "Taxonomy id.",
            "severity": "ERROR",
            "solution": "Rename the subject species to a Latin binomial, obolibrary "
            "taxonomy link, or NCBI taxonomy reference.",
            "source_file_paths": str_nwb_paths,
            "target_file_paths": None,
            "title": "Invalid species",
        },
        {
            "category": "STYLE_SUGGESTION",
            "data_standards": ["BIDS", "DANDI"],
            "examples": ["`ab_01` -> `ab+01`", "`subject #2` -> `subject+2`", "`id 2 from 9/1/25` -> `id+2+9+1+25`"],
            "field": "nwbfile.subject.subject_id",
            "reason": "The participant ID contains invalid characters. BIDS allows only "
            "the plus sign to be used as a separator in the subject entity "
            "label. Underscores, dashes, spaces, slashes, and other special "
            "characters (including #) are expressly forbidden.",
            "severity": "ERROR",
            "solution": "Rename the subject without using any special characters except " "for `+`.",
            "source_file_paths": str_nwb_paths,
            "target_file_paths": None,
            "title": "Invalid participant ID",
        },
        {
            "category": "STYLE_SUGGESTION",
            "data_standards": ["BIDS"],
            "examples": ["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
            "field": "nwbfile.subject.sex",
            "reason": "Participant sex is not one of the allowed patterns by BIDS.",
            "severity": "ERROR",
            "solution": "Rename the subject sex to be one of the accepted values.",
            "source_file_paths": str_nwb_paths,
            "target_file_paths": None,
            "title": "Invalid participant sex (BIDS)",
        },
        {
            "category": "STYLE_SUGGESTION",
            "data_standards": ["DANDI"],
            "examples": ["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
            "field": "nwbfile.subject.sex",
            "reason": "Participant sex is not one of the allowed patterns by the common " "archives.",
            "severity": "ERROR",
            "solution": "Rename the subject sex to be one of the accepted values.",
            "source_file_paths": str_nwb_paths,
            "target_file_paths": None,
            "title": "Invalid participant sex (archives)",
        },
    ]
    assert notifications_json == expected_notification_json


def test_notifications_2(problematic_nwbfile_path_2: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_2]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    notifications = dataset_converter.notifications
    expected_notifications = [
        nwb2bids.Notification(
            title="Missing participant sex",
            reason="Archives such as DANDI or EMBER require the subject sex to be specified.",
            solution=(
                "Specify the `sex` field of the Subject object attached to the NWB file as one of four options: "
                '"M" (for male), "F" (for female), "U" (for unknown), or "O" (for other).\nNOTE: for certain animal '
                'species with more specific genetic determinants, such as C elegans, use "O" (for other) then further '
                'specify the subtypes using other custom fields. For example, `c_elegans_sex="XO"`'
            ),
            examples=None,
            field="nwbfile.subject.sex",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.SCHEMA_INVALIDATION,
            severity=nwb2bids.Severity.CRITICAL,
        ),
        nwb2bids.Notification(
            title="Missing participant species",
            reason="Archives such as DANDI or EMBER require the subject species to be specified.",
            solution=(
                "Specify the `species` field of the Subject object attached to the NWB file as a Latin binomial, "
                "obolibrary taxonomy link, or NCBI taxonomy reference."
            ),
            examples=None,
            field="nwbfile.subject.species",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.SCHEMA_INVALIDATION,
            severity=nwb2bids.Severity.CRITICAL,
        ),
        nwb2bids.Notification(
            title="Invalid participant ID",
            reason=(
                "The participant ID contains invalid characters. BIDS allows only the plus sign to be used as a "
                "separator in the subject entity label. Underscores, dashes, spaces, slashes, and other special "
                "characters (including #) are expressly forbidden."
            ),
            solution="Rename the subject without using any special characters except for `+`.",
            examples=["`ab_01` -> `ab+01`", "`subject #2` -> `subject+2`", "`id 2 from 9/1/25` -> `id+2+9+1+25`"],
            field="nwbfile.subject.subject_id",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.Notification(
            title="Invalid session ID",
            reason=(
                "The session ID contains invalid characters. "
                "BIDS allows only the plus sign to be used as a separator in the subject entity label. "
                "Underscores, dashes, spaces, slashes, and other special characters "
                "(including #) are expressly forbidden."
            ),
            solution="Rename the session without using any special characters except for `+`.",
            examples=[
                "`ses_01` -> `ses+01`",
                "`session #2` -> `session+2`",
                "`id 2 from 9/1/25` -> `id+2+9+1+25`",
            ],
            field="nwbfile.session_id",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
    ]
    assert notifications == expected_notifications


def test_notifications_3(problematic_nwbfile_path_3: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_3]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    notifications = dataset_converter.notifications
    expected_notifications = [
        nwb2bids.Notification(
            title="Missing participant",
            reason="BIDS requires a subject to be specified for each NWB file.",
            solution="Add a Subject object to each NWB file.",
            examples=None,
            field="nwbfile.subject",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.CRITICAL,
        )
    ]
    assert notifications == expected_notifications


def test_notifications_4(problematic_nwbfile_path_4: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_4]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)

    notifications = dataset_converter.notifications
    expected_notifications = [
        nwb2bids.Notification(
            title="Missing description",
            reason="A basic description of this field is recommended to improve contextual understanding.",
            solution="Add a description to the field.",
            examples=None,
            field="nwbfile.devices.DeviceWithoutDescription",
            source_file_paths=None,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.NWB],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.INFO,
        )
    ]
    assert notifications == expected_notifications



"""Configuration file for the Sphinx documentation builder."""

import pathlib
import sys

# Add source directory to path for autodoc
sys.path.insert(0, str(pathlib.Path(__file__).resolve().parents[2] / "src"))

# Project details
project = "nwb2bids"
copyright = "2025, Cody Baker"
author = "Cody Baker"

# Sphinx extensions
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_tabs.tabs",
    "sphinx_copybutton",
    "sphinx_toggleprompt",  # Used to control >>> behavior in the doctests
    "myst_parser",  # For including Markdown files to be rendered as RST
]

# HTML configuration
html_theme = "pydata_sphinx_theme"
html_favicon = "assets/favicon.ico"

html_theme_options = {
    "github_url": "https://github.com/con/nwb2bids",
    "use_edit_page_button": True,
    "show_toc_level": 2,
    "logo": {
        "image_light": "assets/nwb2bids-black.svg",
        "image_dark": "assets/nwb2bids-white.svg",
    }
}

html_context = {
    "github_user": "con",
    "github_repo": "nwb2bids",
    "github_version": "main",
    "doc_path": "docs",
}

html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Format signatures for better readability
autodoc_typehints = "signature"
autodoc_typehints_format = "short"
python_maximum_signature_line_length = 88

# Link checker
linkcheck_anchors = False
linkcheck_ignore = []

# Disable sidebars for specific sections
html_sidebars = {
    'user_guide': [],
    "developer_guide": [],
}

# Toggleprompt
toggleprompt_offset_right = 45  # This controls the position of the prompt (>>>) for the conversion gallery
toggleprompt_default_hidden = "true"

# Copybutton
copybutton_exclude = '.linenos, .gp'  # This avoids copying prompt (>>>) in the conversion gallery (issue #1465)

# MyST
myst_enable_extensions = [
    "colon_fence",      # ::: fences
    "deflist",          # Definition lists
    "fieldlist",        # Field lists
    "html_admonition",  # HTML-style admonitions
    "html_image",       # HTML images
    "replacements",     # Text replacements
    "smartquotes",      # Smart quotes
    "strikethrough",    # ~~strikethrough~~
    "substitution",     # Variable substitutions
    "tasklist",         # Task lists
]
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}


# --------------------------------------------------
# Extension configuration
# --------------------------------------------------

# Napoleon
napoleon_google_docstring = False          # Disable support for Google-style docstrings (use NumPy-style instead)
napoleon_numpy_docstring = True            # Enable support for NumPy-style docstrings
napoleon_use_param = False                 # Do not convert :param: sections into Parameters; leave as-is
napoleon_use_ivar = True                   # Interpret instance variables as documented with :ivar:

# Autodoc
autoclass_content = "both"
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "private-members": False,
    "undoc-members": True,
    "show-inheritance": False,
    "exclude-members": "__new__,model_dump",
}
add_module_names = False
toc_object_entries_show_parents = "hide"

def setup(app) -> None:
    app.connect("html-page-context", remove_section_nav)

def remove_section_nav(app, pagename, templatename, context, doctree) -> None:
    if doctree:
        # Count the number of top-level sections
        sections = [node for node in doctree.traverse() if node.tagname == "section"]
        if len(sections) <= 1:
            # Remove the 'sidebar-secondary' if it exists
            context["sidebars"] = [sidebar for sidebar in context.get("sidebars", []) if sidebar != "sidebar-secondary"]



"""Configuration file for sybil-based doc testing."""
import json
import pathlib
import shutil
import subprocess

from sybil import Sybil
from sybil.parsers.rest import CodeBlockParser, PythonCodeBlockParser, SkipParser

import nwb2bids


def bash_evaluator(example):
    """Execute bash code blocks."""
    result = subprocess.run(
        example.parsed,
        shell=True,
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        return f"Command failed (exit {result.returncode}):\n{result.stderr}"


def sybil_setup(namespace):
    """Clean and regenerate tutorial data before tests.

    NOTE: This runs once per document, not per code block. If multiple examples
    in the same document write to the same output directory, they will conflict.
    Use distinct output directories (e.g., bids_dataset_cli_1, bids_dataset_py_1).
    """
    tutorial_base = nwb2bids.testing.get_tutorial_directory()
    if tutorial_base.exists():
        shutil.rmtree(tutorial_base)

    nwb2bids.testing.generate_ephys_tutorial(mode="file")
    nwb2bids.testing.generate_ephys_tutorial(mode="dataset")

    # Create metadata.json for tutorials that need it
    tutorial_dir = tutorial_base / "ecephys_tutorial_file"
    (tutorial_dir / "metadata.json").write_text(json.dumps({
        "dataset_description": {
            "Name": "My Custom BIDS Dataset",
            "BIDSVersion": "1.8.0",
            "HEDVersion": "8.3.0",
            "Authors": ["First Last", "Second Author"]
        }
    }))

    # Make common imports and paths available
    namespace["Path"] = pathlib.Path
    namespace["pathlib"] = pathlib
    namespace["nwb2bids"] = nwb2bids
    namespace["tutorial_base"] = tutorial_base


pytest_collect_file = Sybil(
    parsers=[
        SkipParser(),
        CodeBlockParser(language="bash", evaluator=bash_evaluator),
        PythonCodeBlockParser(),
    ],
    patterns=["tutorials.rst"],
    setup=sybil_setup,
).pytest()



import os
import pathlib
import sys

import pytest

import nwb2bids

# These tests fail on Windows GitHub CI due to git-annex adjusted branch issues
# See https://github.com/con/nwb2bids/pull/213 for failure output
pytest_mark_xfail_windows_github_ci = pytest.mark.xfail(
    sys.platform == "win32" and os.environ.get("GITHUB_ACTIONS", "").lower() == "true",
    reason="git-annex adjusted branch fails on Windows GitHub CI runners",
    strict=False,
)


@pytest.mark.remote
def test_remote_convert_nwb_dataset(temporary_bids_directory: pathlib.Path):
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.DatasetConverter.from_remote_dandiset(
        dandiset_id="000003", limit=2, run_config=run_config
    )
    dataset_converter.extract_metadata()
    dataset_converter.convert_to_bids_dataset()
    assert len(dataset_converter.notifications) == 4

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-YutaMouse20"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-YutaMouse20": {
            "directories": {"ses-YutaMouse20-140321", "ses-YutaMouse20-140327"},
            "files": {"sub-YutaMouse20_sessions.json", "sub-YutaMouse20_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140321": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140327": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140321"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-YutaMouse20_ses-YutaMouse20-140321_channels.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_channels.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140321_ecephys.nwb",
                "sub-YutaMouse20_ses-YutaMouse20-140321_electrodes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_electrodes.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140321_probes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_probes.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140321_events.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_events.tsv",
            },
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140327"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-YutaMouse20_ses-YutaMouse20-140327_channels.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_channels.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140327_ecephys.nwb",
                "sub-YutaMouse20_ses-YutaMouse20-140327_electrodes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_electrodes.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140327_probes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_probes.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140327_events.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_events.tsv",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


@pytest.mark.remote
@pytest_mark_xfail_windows_github_ci
def test_remote_convert_nwb_dataset_on_gotten_datalad_file(
    testing_files_directory: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    import datalad.api

    dataset_dir = testing_files_directory / "000568"
    dataset_dir.mkdir(exist_ok=True)

    dataset = datalad.api.clone(source="https://github.com/dandisets/000568", path=dataset_dir)

    filename = "sub-fCamk1_ses-fCamk1-200827-sess9-no-raw-data_behavior+ecephys+image+ogen.nwb"
    test_file_path = dataset_dir / "sub-fCamk1" / filename
    dataset.get(path=test_file_path)

    nwb_paths = [test_file_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert len(dataset_converter.notifications) < 2, "Expected fewer than 2 notifications!"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-fCamk1"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-fCamk1": {
            "directories": {"ses-fCamk1_200827_sess9_no_raw_data"},
            "files": {"sub-fCamk1_sessions.json", "sub-fCamk1_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-fCamk1"
        / "ses-fCamk1_200827_sess9_no_raw_data": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-fCamk1"
        / "ses-fCamk1_200827_sess9_no_raw_data"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_channels.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_channels.tsv",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_ecephys.nwb",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_electrodes.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_electrodes.tsv",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_probes.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_probes.tsv",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_events.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_events.tsv",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


@pytest.mark.remote
@pytest_mark_xfail_windows_github_ci
def test_remote_convert_nwb_dataset_on_partial_datalad_dataset(
    testing_files_directory: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    import datalad.api

    dataset_dir = testing_files_directory / "000568"
    dataset_dir.mkdir(exist_ok=True)

    dataset = datalad.api.clone(source="https://github.com/dandisets/000568", path=dataset_dir)

    filename = "sub-fCamk1_ses-fCamk1-200827-sess9-no-raw-data_behavior+ecephys+image+ogen.nwb"
    test_file_path = dataset_dir / "sub-fCamk1" / filename
    dataset.get(path=test_file_path)

    nwb_paths = [dataset_dir]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert len(dataset_converter.notifications) < 2, "Expected fewer than 2 notifications!"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-fCamk1"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-fCamk1": {
            "directories": {"ses-fCamk1_200827_sess9_no_raw_data"},
            "files": {"sub-fCamk1_sessions.json", "sub-fCamk1_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-fCamk1"
        / "ses-fCamk1_200827_sess9_no_raw_data": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-fCamk1"
        / "ses-fCamk1_200827_sess9_no_raw_data"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_channels.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_channels.tsv",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_ecephys.nwb",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_electrodes.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_electrodes.tsv",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_probes.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_probes.tsv",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_events.json",
                "sub-fCamk1_ses-fCamk1_200827_sess9_no_raw_data_events.tsv",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import json
import pathlib


def _validate_bids_directory(directory: pathlib.Path) -> pathlib.Path:
    """Validate bids_directory: if exists, must be valid BIDS; if not, parent must exist."""
    if directory.exists():
        return _validate_existing_directory_as_bids(directory)

    if not directory.parent.exists():
        raise ValueError(f"parent directory does not exist: {directory.parent}")

    return directory


def _validate_existing_directory_as_bids(directory: pathlib.Path) -> pathlib.Path:

    dataset_description_file_path = directory / "dataset_description.json"
    current_directory_contents = {path.stem for path in directory.iterdir() if not path.name.startswith(".")} - {
        "README",
        "CHANGES",
        "derivatives",
        "dandiset",
    }

    if len(current_directory_contents) == 0:
        # The directory is considered empty, not containing any meaningful content.
        # Populate the directory with `dataset_description.json` to make it
        # a valid (though minimal) BIDS dataset.

        default_dataset_description = {"BIDSVersion": "1.10", "HEDVersion": "8.3.0"}
        with dataset_description_file_path.open(mode="w") as file_stream:
            json.dump(obj=default_dataset_description, fp=file_stream, indent=4)
    else:
        # The directory is considered non-empty

        if not dataset_description_file_path.is_file():
            # The directory is without `dataset_description.json` file
            # It is an invalid BIDS dataset.

            message = (
                f"The directory ({directory}) exists and is not empty, but is not a valid BIDS dataset: "
                "missing 'dataset_description.json' file."
            )
            raise ValueError(message)

        with dataset_description_file_path.open(mode="r") as file_stream:
            try:
                dataset_description = json.load(fp=file_stream)
            except json.JSONDecodeError as exception:
                message = (
                    f"The directory ({directory}) exists and contains a 'dataset_description.json' file, "
                    "but it is not a valid JSON file."
                )
                raise ValueError(message) from exception

        if not isinstance(dataset_description, dict):
            message = (
                f"The directory ({directory}) exists and contains a 'dataset_description.json' file, "
                "but it does not contain a valid JSON object."
            )
            raise ValueError(message)
        if dataset_description.get("BIDSVersion", None) is None:
            message = (
                f"The directory ({directory}) exists but is not a valid BIDS dataset: "
                "missing 'BIDSVersion' in 'dataset_description.json'."
            )
            raise ValueError(message)

    return directory



"""Sanitization tests for the API."""

import pathlib

import nwb2bids


def test_convert_nwb_dataset_basic_sanitization(
    problematic_nwbfile_path_2: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [problematic_nwbfile_path_2]
    run_config = nwb2bids.RunConfig(
        bids_directory=temporary_bids_directory,
        sanitization_config=nwb2bids.sanitization.SanitizationConfig(sub_labels=True, ses_labels=True),
    )
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert len(dataset_converter.notifications) == 4

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-bad+subject+id"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-bad+subject+id": {
            "directories": {"ses-problematic+2"},
            "files": {"sub-bad+subject+id_sessions.json", "sub-bad+subject+id_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-bad+subject+id"
        / "ses-problematic+2": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-bad+subject+id"
        / "ses-problematic+2"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-bad+subject+id_ses-problematic+2_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import pathlib
import tempfile
import typing


def _determine_file_mode() -> typing.Literal["symlink", "copy"]:
    """
    Determine what file mode to use in creating a BIDS dataset based on the system
    """
    with tempfile.TemporaryDirectory(prefix="nwb2bids-") as temp_dir_str:
        temp_dir_path = pathlib.Path(temp_dir_str)

        # Create a test file to determine if symlinks are supported
        test_file_path = temp_dir_path / "test_file.txt"
        test_file_path.touch()

        try:
            # Create a symlink to the test file
            (temp_dir_path / "test_symlink.txt").symlink_to(target=test_file_path)
        except (OSError, PermissionError, NotImplementedError):  # Windows can sometimes have trouble with symlinks
            # TODO: log a INFO message here when logging is set up
            return "copy"
        else:
            # If symlink creation was successful, return "symlink"
            return "symlink"



import pathlib

import pydantic
import pytest

import nwb2bids


def test_run_config_immutability(temporary_bids_directory: pathlib.Path):
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)

    with pytest.raises(expected_exception=pydantic.ValidationError, match="Instance is frozen"):
        run_config.run_id = "new_run_id"


def test_run_config_accepts_nonexistent_bids_directory(temporary_bids_directory: pathlib.Path):
    """Test that RunConfig accepts a nonexistent bids_directory if parent exists (but doesn't create it)."""
    nonexistent_child = temporary_bids_directory / "new_bids_dir"

    assert not nonexistent_child.exists()
    run_config = nwb2bids.RunConfig(bids_directory=nonexistent_child)
    assert not nonexistent_child.exists()  # NOT created at init time
    assert run_config.bids_directory == nonexistent_child


def test_run_config_missing_parent_raises(temporary_bids_directory: pathlib.Path):
    """Test that an exception is raised when parent directory doesn't exist."""
    nested_bids_directory = temporary_bids_directory / "nonexistent_parent" / "child"

    with pytest.raises(expected_exception=pydantic.ValidationError, match="parent directory does not exist"):
        nwb2bids.RunConfig(bids_directory=nested_bids_directory)



import collections
import pathlib
import typing
import warnings

import rich_click

from .._converters._run_config import RunConfig
from .._core._convert_nwb_dataset import convert_nwb_dataset
from .._tools._pluralize import _pluralize
from ..notifications import Notification, Severity
from ..sanitization import SanitizationConfig
from ..testing import generate_ephys_tutorial


# nwb2bids
@rich_click.group(name="nwb2bids")
def _nwb2bids_cli():
    pass


# nwb2bids convert [OPTIONS] [NWB_PATHS]...
@_nwb2bids_cli.command(name="convert")
@rich_click.argument("nwb_paths", type=str, nargs=-1)
@rich_click.option(
    "--bids-directory",
    "-o",
    help="Path to the folder where the BIDS dataset will be created (default: current working directory).",
    required=False,
    type=rich_click.Path(writable=True),
    default=None,
)
@rich_click.option(
    "--file-mode",
    help=(
        "How to handle the source NWB files when converting to BIDS format. "
        "By default, will attempt to utilize 'symlink' if the system allows. "
        "Otherwise, will 'copy' to preserve directory integrity. "
        "Use 'move' for fastest speeds if you do not wish to keep the original NWB directory structure."
    ),
    required=False,
    type=rich_click.Choice(["copy", "move", "symlink", "auto"], case_sensitive=False),
    default="auto",
)
@rich_click.option(
    "--cache-directory",
    "cache_directory",
    help=(
        "The directory where run specific files (e.g., notifications, sanitization reports) will be stored. "
        "Defaults to `~/.nwb2bids`."
    ),
    required=False,
    type=rich_click.Path(exists=True, dir_okay=False, readable=True),
    default=None,
)
@rich_click.option(
    "--additional-metadata-file-path",
    "additional_metadata_file_path",
    help=(
        "Path to a JSON file containing additional metadata to be included in the BIDS dataset. "
        "This file should contain a dictionary with keys corresponding to BIDS entities."
    ),
    required=False,
    type=rich_click.Path(exists=True, dir_okay=False, readable=True),
    default=None,
)
@rich_click.option(
    "--sanitization",
    help="Specifies types of sanitization to apply when creating the BIDS dataset.",
    required=False,
    type=rich_click.Choice(["sub-labels", "ses-labels"], case_sensitive=True),
    multiple=True,
    default=None,
)
@rich_click.option("--silent", "-s", is_flag=True, help="Suppress all console output.", default=False)
@rich_click.option(
    "--run-id",
    help=(
        "On each unique run of nwb2bids, a run ID is generated. "
        "Set this option to override this to any identifying string. "
        "This ID is used in the naming of the notification and sanitization reports saved to your cache directory. "
        'The default ID uses runtime timestamp information of the form "date-%Y%m%d_time-%H%M%S."'
    ),
    required=False,
    type=str,
    default=None,
)
def _run_convert_nwb_dataset(
    nwb_paths: tuple[str, ...],
    bids_directory: str | None = None,
    sanitization: tuple[typing.Literal["sub-labels", "ses-labels"]] = (),
    additional_metadata_file_path: str | None = None,
    file_mode: typing.Literal["copy", "move", "symlink", "auto"] = "auto",
    cache_directory: str | None = None,
    run_id: str | None = None,
    silent: bool = False,
) -> None:
    """
    Convert NWB files to BIDS format.

    NWB_PATHS : A space-separated sequence of paths, each pointing to either an NWB file
    or a directory containing NWB files.
    """
    if silent:
        warnings.filterwarnings(action="ignore")

    if len(nwb_paths) == 0:
        message = "Please provide at least one NWB file or directory to convert."
        raise ValueError(message)
    handled_nwb_paths = [pathlib.Path(nwb_path) for nwb_path in nwb_paths]
    # Convert CLI args to snake_case
    sanitization_config = SanitizationConfig(**{value.replace("-", "_"): True for value in sanitization})

    run_config_kwargs = {
        "bids_directory": bids_directory,
        "additional_metadata_file_path": additional_metadata_file_path,
        "file_mode": file_mode,
        "cache_directory": cache_directory,
        "sanitization_config": sanitization_config,
        "run_id": run_id,
    }

    # Filter out values that indicate absence of direct user input or signal to use default
    non_missing_run_config_kwargs = {
        key: value
        for key, value in run_config_kwargs.items()
        if (key != "file_mode" and value is not None) or (key == "file_mode" and value != "auto")
    }
    run_config = RunConfig(**non_missing_run_config_kwargs)

    dataset_converter = convert_nwb_dataset(nwb_paths=handled_nwb_paths, run_config=run_config)

    if silent:
        return

    notifications = dataset_converter.notifications
    notifications_by_severity: dict[Severity, list[Notification]] = collections.defaultdict(list)
    for notification in notifications:
        notifications_by_severity[notification.severity].append(notification)
    notif_text = f"\n\nPlease review the full notifications report at {run_config.notifications_json_file_path}\n"

    errors = notifications_by_severity[Severity.ERROR]
    criticals = notifications_by_severity[Severity.CRITICAL]

    if errors:
        number_of_errors = len(errors)

        top_three = errors[:3]
        number_to_print = len(top_three)

        text = (
            "\nBIDS dataset was not successfully created!\n"
            f'{_pluralize(n=number_to_print, phrase="An error was", plural="Some errors were")} '
            "encountered during conversion.\n"
        )
        error_text = "".join(f"\n\t- {error.reason}" for error in top_three)
        if number_to_print > 1 and number_of_errors > 3:
            counting_text = f"The first {number_to_print} of {number_of_errors} are shown below:"
        elif number_to_print >= 2:
            counting_text = f"The first {number_to_print} are shown below:"
        else:
            counting_text = "The error is shown below:"
        text += f"{counting_text}\n\n{error_text}{notif_text}"

        console_notification = rich_click.style(text=text, fg="red")
        rich_click.echo(message=console_notification)
        return

    sanitization_text = ""
    if any(sanitization_config.model_dump().values()):
        sanitization_text = (
            "\n\nNote: Sanitization was applied to file and directory names during conversion. "
            "Please review the converted BIDS dataset to ensure all names are appropriate.\n\n"
        )

    if criticals:
        text = f"\nBIDS dataset was successfully created, but may not be valid!{sanitization_text}{notif_text}"
        console_notification = rich_click.style(text=text, fg="yellow")
        rich_click.echo(message=console_notification)
        return

    text = "\nBIDS dataset was successfully created!\n"
    if notifications:
        number_of_notifications = len(notifications)

        text += (
            f'{number_of_notifications} {_pluralize(n=number_of_notifications, phrase="suggestion")} for improvement '
            f'{_pluralize(n=number_of_notifications, phrase="was", plural="were")} found during conversion.'
            f"{sanitization_text}{notif_text}"
        )
    console_notification = rich_click.style(text=text, fg="green")
    rich_click.echo(message=console_notification)


# nwb2bids tutorial
@_nwb2bids_cli.group(name="tutorial")
def _nwb2bids_tutorial_cli():
    pass


# nwb2bids tutorial ephys
@_nwb2bids_tutorial_cli.group(name="ephys")
def _nwb2bids_tutorial_ephys_cli():
    pass


# nwb2bids tutorial ephys file
@_nwb2bids_tutorial_ephys_cli.command(name="file")
@rich_click.option(
    "--output-directory",
    "-o",
    help="Path to the folder where the tutorial file(s) will be created (default: user home directory).",
    required=False,
    type=rich_click.Path(writable=True),
    default=None,
)
@rich_click.option(
    "--modality",
    help="Type of electrophysiology data to include in the tutorial NWB file.",
    required=False,
    type=rich_click.Choice(["ecephys", "icephys"], case_sensitive=True),
    default="ecephys",
)
def _nwb2bids_tutorial_ephys_file_cli(
    output_directory: str | None = None,
    modality: typing.Literal["ecephys", "icephys"] = "ecephys",
) -> None:
    file_path = generate_ephys_tutorial(output_directory=output_directory, mode="file", modality=modality)

    text = f"\nAn example NWB file has been created at: {file_path}\n\n"
    message = rich_click.style(text=text, fg="green")
    rich_click.echo(message=message)


# nwb2bids tutorial ephys dataset
@_nwb2bids_tutorial_ephys_cli.command(name="dataset")
@rich_click.option(
    "--output-directory",
    "-o",
    help="Path to the folder where the tutorial files will be created (default: user home directory).",
    required=False,
    type=rich_click.Path(writable=True),
    default=None,
)
@rich_click.option(
    "--modality",
    help="Type of electrophysiology data to include in the tutorial NWB file.",
    required=False,
    type=rich_click.Choice(["ecephys", "icephys"], case_sensitive=True),
    default="ecephys",
)
def _nwb2bids_tutorial_ephys_dataset_cli(
    output_directory: str | None = None,
    modality: typing.Literal["ecephys", "icephys"] = "ecephys",
) -> None:
    tutorial_directory = generate_ephys_tutorial(output_directory=output_directory, mode="dataset", modality=modality)

    text = f"\nAn example NWB dataset has been created at: {tutorial_directory}\n\n"
    message = rich_click.style(text=text, fg="green")
    rich_click.echo(message=message)



import abc

import pydantic

from ..notifications import Notification


class MutableModel(pydantic.BaseModel):
    """
    Base Pydantic model for all mutable models in `nwb2bids`.

    Parameters
    ----------
    _file_paths : list of str
        Private attribute to store file paths associated with this model.
        Used by the message feature to associate particular issues with particular files.
    """

    model_config = pydantic.ConfigDict(
        validate_assignment=True,  # Re-validate model on mutation
        extra="allow",  # Allow additional custom fields
    )


class BaseMetadataModel(MutableModel):
    """
    Base Pydantic model for all metadata handled by `nwb2bids`.
    """

    notifications: list[Notification] = pydantic.Field(
        description="List of auto-detected notifications.",
        default_factory=list,
        exclude=True,
    )


class BaseMetadataContainerModel(MutableModel, abc.ABC):
    """
    Base Pydantic model for 'containing' some number of other metadata models plus any extra metadata.
    """

    _internal_notifications: list[Notification] = pydantic.PrivateAttr(default_factory=list)

    @abc.abstractmethod
    @pydantic.computed_field
    @property
    def notifications(self) -> list[Notification]:
        """
        All notifications from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        message = "This field should be defined in child classes."
        raise NotImplementedError(message)



"""Sanitization tests for the primary `nwb2bids convert` CLI."""

import pathlib
import subprocess

import nwb2bids


def test_cli_basic_sanitization(problematic_nwbfile_path_2: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = (
        f"nwb2bids convert {problematic_nwbfile_path_2} -o {temporary_bids_directory} "
        "--sanitization sub-labels --sanitization ses-labels"
    )

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert (
        result.returncode == 0
    ), f"\n\nCLI command failed with:\nStandard Output: {result.stdout}\nStandard Error: {result.stderr}\n\n"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-bad+subject+id"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-bad+subject+id": {
            "directories": {"ses-problematic+2"},
            "files": {"sub-bad+subject+id_sessions.json", "sub-bad+subject+id_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-bad+subject+id"
        / "ses-problematic+2": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-bad+subject+id"
        / "ses-problematic+2"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-bad+subject+id_ses-problematic+2_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import collections
import json
import traceback

import pandas
import pydantic
import typing_extensions

from ._dandi_utils import get_bids_dataset_description
from ._run_config import RunConfig
from ._session_converter import SessionConverter
from .._converters._base_converter import BaseConverter
from ..bids_models import BidsSessionMetadata, DatasetDescription
from ..notifications import Category, Notification, Severity


class DatasetConverter(BaseConverter):
    session_converters: list[SessionConverter] = pydantic.Field(
        description="List of session converters. Typically instantiated by calling `.from_nwb_paths()`."
    )
    dataset_description: DatasetDescription | None = pydantic.Field(
        description="The BIDS-compatible dataset description.",
        default=None,
    )

    @pydantic.computed_field
    @property
    def notifications(self) -> list[Notification]:
        """
        All notifications from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        notifications = [
            notification
            for session_converter in self.session_converters
            for notification in session_converter.notifications
        ] + self._internal_notifications
        notifications.sort(
            key=lambda notification: (-notification.category.value, -notification.severity.value, notification.title)
        )
        return notifications

    @classmethod
    @pydantic.validate_call
    def from_remote_dandiset(
        cls,
        dandiset_id: str = pydantic.Field(pattern=r"^\d{6}$"),
        api_url: str | None = None,
        token: str | None = None,
        limit: int | None = None,
        run_config: RunConfig = pydantic.Field(default_factory=lambda: RunConfig()),
    ) -> typing_extensions.Self | None:
        """
        Initialize a converter of a Dandiset to BIDS format.

        Parameters
        ----------
        dandiset_id : str
            The dandiset ID of the Dandiset to be converted.
        api_url : str, optional
            The API URL of a custom DANDI instance to use. If not provided, the API URL of the
            DANDI instance specified by the :envvar:`DANDI_INSTANCE` environment variable
            is used. If the :envvar:`DANDI_INSTANCE` environment variable is not specified,
            The API URL of the `"dandi"` DANDI instance is used.
        token : str, optional
            The authentication token for accessing the DANDI instance.
            If not provided, will attempt to read from the environment variable `DANDI_API_KEY` if it exists.
            This is required for accessing embargoed Dandisets.
        limit : int, optional
            If specified, limits the number of sessions to convert.
            This is mainly useful for testing purposes.
        run_config : RunConfig, optional
            The configuration for this conversion run.
        """
        try:
            import dandi.dandiapi

            client = dandi.dandiapi.DandiAPIClient(api_url=api_url, token=token)
            version_id = "draft"  # Only allow running on draft version
            dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=version_id)

            dataset_description, _internal_notifications = get_bids_dataset_description(dandiset=dandiset)

            if limit is None:
                assets = list(dandiset.get_assets())
            else:
                assets = [asset for counter, asset in enumerate(dandiset.get_assets()) if counter < limit]

            session_id_to_assets = collections.defaultdict(list)
            for asset in assets:
                asset_metadata = asset.get_raw_metadata()

                for session in asset_metadata.get("wasGeneratedBy", []):
                    if session.get("schemaKey", "") != "Session":
                        continue

                    session_id = session.get("identifier", "")
                    if session_id == "":
                        continue

                    session_id_to_assets[session_id].append(asset)
            sorted_session_id_to_assets = dict(sorted(session_id_to_assets.items(), key=lambda item: item[0]))

            session_converters = [
                SessionConverter(
                    session_id=session_id,
                    nwbfile_paths=[asset.get_content_url(follow_redirects=1, strip_query=True) for asset in assets],
                    run_config=run_config,
                )
                for session_id, assets in sorted_session_id_to_assets.items()
            ]

            dataset_converter = cls(
                session_converters=session_converters, dataset_description=dataset_description, run_config=run_config
            )
            dataset_converter._internal_notifications = _internal_notifications
            return dataset_converter
        except Exception:  # noqa
            _internal_notifications = [
                Notification(
                    title="Failed to initialize converter on remote Dandiset",
                    reason=(
                        "An error occurred while executing `DatasetConverter.from_remote_dandiset`."
                        f"\n\n{traceback.format_exc()}"
                    ),
                    solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                    category=Category.INTERNAL_ERROR,
                    severity=Severity.ERROR,
                )
            ]

            dataset_converter = cls(session_converters=[], dataset_description=None, run_config=run_config)

        dataset_converter._internal_notifications = _internal_notifications
        return dataset_converter

    @classmethod
    @pydantic.validate_call
    def from_nwb_paths(
        cls,
        nwb_paths: list[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
        run_config: RunConfig = pydantic.Field(default_factory=lambda: RunConfig()),
    ) -> typing_extensions.Self:
        """
        Initialize a converter of NWB files to BIDS format.

        Parameters
        ----------
        nwb_paths : iterable of file and directory paths
            An iterable of NWB file paths and directories containing NWB files.
        run_config : RunConfig, optional
            The configuration for this conversion run.

        Returns
        -------
        An instance of DatasetConverter.
        """
        try:
            session_converters = SessionConverter.from_nwb_paths(nwb_paths=nwb_paths, run_config=run_config)

            dataset_description = None
            additional_metadata_file_path = run_config.additional_metadata_file_path
            if additional_metadata_file_path is not None:
                dataset_description = DatasetDescription.from_file_path(file_path=additional_metadata_file_path)

            session_messages = [
                notification
                for session_converter in session_converters
                for notification in session_converter.notifications
            ]

            dataset_converter = cls(
                session_converters=session_converters, dataset_description=dataset_description, run_config=run_config
            )
            dataset_converter._internal_notifications = session_messages
            return dataset_converter
        except Exception:  # noqa
            _internal_notifications = [
                Notification(
                    title="Failed to initialize converter on local NWB files",
                    reason=(
                        "An error occurred while executing `DatasetConverter.from_nwb_paths`."
                        f"\n\n{traceback.format_exc()}"
                    ),
                    solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                    category=Category.INTERNAL_ERROR,
                    severity=Severity.ERROR,
                )
            ]
            dataset_converter = cls(session_converters=[], dataset_description=None, run_config=run_config)
            dataset_converter._internal_notifications = _internal_notifications
            return dataset_converter

    def extract_metadata(self) -> None:
        try:
            collections.deque(
                (
                    session_converter.extract_metadata()
                    for session_converter in self.session_converters
                    if session_converter.session_metadata is None
                ),
                maxlen=0,
            )
        except Exception:  # noqa
            notification = Notification(
                title="Failed to extract metadata for one or more sessions",
                reason=(
                    "An error occurred while executing `DatasetConverter.extract_metadata`."
                    f"\n\n{traceback.format_exc()}"
                ),
                solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                category=Category.INTERNAL_ERROR,
                severity=Severity.ERROR,
            )
            self._internal_notifications.append(notification)

    def convert_to_bids_dataset(self) -> None:
        """Convert the directory of NWB files to a BIDS dataset."""
        try:
            for session_converter in self.session_converters:
                session_converter.convert_to_bids_session()

            self.write_participants_metadata()
            self.write_sessions_metadata()
            self.write_dataset_description()
        except Exception:  # noqa
            notification = Notification(
                title="Failed to convert to BIDS dataset",
                reason=(
                    "An error occurred while executing `DatasetConverter.convert_to_bids_dataset`."
                    f"\n\n{traceback.format_exc()}"
                ),
                solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                category=Category.INTERNAL_ERROR,
                severity=Severity.ERROR,
            )
            self._internal_notifications.append(notification)
        finally:
            self.run_config.bids_directory.mkdir(exist_ok=True)  # Just in case it failed to create earlier
            self.run_config._nwb2bids_directory.mkdir(exist_ok=True)
            notifications_dump = [notification.model_dump(mode="json") for notification in self.notifications]
            self.run_config.notifications_json_file_path.write_text(data=json.dumps(obj=notifications_dump, indent=2))

    def write_dataset_description(self) -> None:
        """Write the `dataset_description.json` file."""
        if self.dataset_description is None:
            self.dataset_description = DatasetDescription(BIDSVersion="1.10.1", HEDVersion="8.3.0")

        dataset_description_dictionary = self.dataset_description.model_dump()

        dataset_description_file_path = self.run_config.bids_directory / "dataset_description.json"
        with dataset_description_file_path.open(mode="w") as file_stream:
            json.dump(obj=dataset_description_dictionary, fp=file_stream, indent=4)

    def write_participants_metadata(self) -> None:
        """Write the `participants.tsv` and `participants.json` files."""
        model_dump_per_session = []
        for session_converter in self.session_converters:
            model_dump = session_converter.session_metadata.participant.model_dump()
            model_dump_per_session.append(model_dump)

        full_participants_data_frame = pandas.DataFrame.from_records(
            data=[
                {key: value for key, value in model_dump.items() if value is not None}
                for model_dump in model_dump_per_session
            ]
        ).astype("string")

        if full_participants_data_frame.empty:
            return

        # Aggregate values across entries (such as species mismatches)
        cols_to_agg = [col for col in full_participants_data_frame.columns if col != "participant_id"]
        if any(cols_to_agg):
            aggregated_cols = {col: lambda val: ", ".join(val.dropna().unique()) for col in cols_to_agg}
            aggregated_data_frame = full_participants_data_frame.groupby(by="participant_id", as_index=False).agg(
                func=aggregated_cols
            )
        else:
            aggregated_data_frame = full_participants_data_frame.copy()

        # Deduplicate all rows of the frame
        deduplicated_data_frame = aggregated_data_frame.drop_duplicates(ignore_index=True).copy()

        # Apply sanitization
        sanitizations = [converter.session_metadata.sanitization for converter in self.session_converters]
        sanitized_participant_ids = {
            sanitization.original_participant_id: sanitization.sanitized_participant_id
            for sanitization in sanitizations
        }

        with pandas.option_context("mode.chained_assignment", None):
            deduplicated_data_frame["participant_id"] = (
                deduplicated_data_frame["participant_id"]
                .apply(lambda participant_id: sanitized_participant_ids[participant_id])
                .astype("string")
            )

        # BIDS requires sub- prefix in table values
        participants_data_frame = deduplicated_data_frame.copy(deep=True)
        participants_data_frame["participant_id"] = (
            participants_data_frame["participant_id"]
            .apply(lambda participant_id: f"sub-{participant_id}")
            .astype("string")
        )
        is_field_in_table = {field: True for field in participants_data_frame.keys()}

        # BIDS Validator is strict regarding column order
        required_column_order = [
            field
            for field in ["participant_id", "species", "sex", "strain"]
            if is_field_in_table.get(field, False) is True
        ]
        column_order = required_column_order + [
            field
            for field in participants_data_frame.columns
            if is_field_in_table.get(field, False) is True and field not in required_column_order
        ]

        participants_tsv_file_path = self.run_config.bids_directory / "participants.tsv"
        participants_data_frame.to_csv(
            path_or_buf=participants_tsv_file_path, mode="w", index=False, sep="\t", columns=column_order
        )

        if len(self.session_converters) > 0:
            is_field_in_table = {field: True for field in participants_data_frame.keys()}
            participants_schema = self.session_converters[0].session_metadata.participant.model_json_schema()
            participants_json = {
                field: info["description"]
                for field, info in participants_schema["properties"].items()
                if is_field_in_table.get(field, False) is True
            }
            participants_json_file_path = self.run_config.bids_directory / "participants.json"
            with participants_json_file_path.open(mode="w") as file_stream:
                json.dump(obj=participants_json, fp=file_stream, indent=4)

    def write_sessions_metadata(self) -> None:
        """
        Write the `_sessions.tsv` and `_sessions.json` files, then create empty participant and session directories.
        """
        participant_id_to_sessions = collections.defaultdict(list)
        for session_converter in self.session_converters:
            participant_id_to_sessions[session_converter.session_metadata.sanitization.sanitized_participant_id].append(
                session_converter.session_metadata
            )

        # TODO: expand beyond just session_id field (mainly via additional metadata)
        sessions_schema = BidsSessionMetadata.model_json_schema()
        sessions_json = {"session_id": sessions_schema["properties"]["session_id"]["description"]}

        for participant_id, sessions_metadata in participant_id_to_sessions.items():
            sanitized_participant_id = participant_id
            sanitized_session_ids = [
                session_metadata.sanitization.sanitized_session_id for session_metadata in sessions_metadata
            ]

            subject_directory = self.run_config.bids_directory / f"sub-{sanitized_participant_id}"
            subject_directory.mkdir(exist_ok=True)

            # BIDS requires ses- prefix in table values
            sessions_data_frame = pandas.DataFrame(
                {"session_id": [f"ses-{session_id}" for session_id in sanitized_session_ids]}
            )

            session_tsv_file_path = subject_directory / f"sub-{sanitized_participant_id}_sessions.tsv"
            sessions_data_frame.to_csv(path_or_buf=session_tsv_file_path, mode="w", index=False, sep="\t")

            session_json_file_path = subject_directory / f"sub-{sanitized_participant_id}_sessions.json"
            with session_json_file_path.open(mode="w") as file_stream:
                json.dump(obj=sessions_json, fp=file_stream, indent=4)

            for session_id in sanitized_session_ids:
                session_directory = subject_directory / f"ses-{session_id}"
                session_directory.mkdir(exist_ok=True)



import pydantic


class SanitizationConfig(pydantic.BaseModel):
    """
    The different types of sanitization that can be applied to the `nwb2bids` conversion process.

    Not enabling many of these options may lead to the creation of an invalid BIDS dataset.

    Attributes
    ----------
    sub_labels : bool, default: False
        Subject labels are sanitized by replacing any non-alphanumeric characters with plus signs.
    ses_labels : bool, default: False
        Session labels are sanitized by replacing any non-alphanumeric characters with plus signs.

    Examples
    --------
    A file such as `nwbfile.subject.subject_id = "My Subject_01"` becomes the `sub-My+Subject+01` entity-label pair
    in BIDS filenames when using `SUB_LABELS = True`.
    Otherwise, the filenames would have included the string `sub-My Subject_01`.
    """

    sub_labels: bool = False
    ses_labels: bool = False

    model_config = pydantic.ConfigDict(
        frozen=True,  # Make the model immutable
        validate_default=True,  # Validate default values as well
    )



import pathlib

import py.path

import nwb2bids


def test_ephys_tutorial_generation(tmpdir: py.path.local):
    tutorial_path = nwb2bids.testing.generate_ephys_tutorial(output_directory=pathlib.Path(tmpdir), mode="file")

    assert tutorial_path.exists()
    assert tutorial_path.suffix == ".nwb"



import pathlib
import re
import typing

import pydantic

from ._configuration import SanitizationConfig


class Sanitization(pydantic.BaseModel):
    """
    Sanitize all relevant metadata fields when translating from minimal NWB towards valid BIDS.

    sanitization_config : SanitizationConfig
        The types of sanitization to apply.
        See `nwb2bids.sanitization.SanitizationConfig?` for more details.
    sanitization_file_path : pathlib.Path
        The file to save a record of sanitization actions taken.
    original_session_id : str
        The original session ID to be sanitized.
    original_participant_id : str
        The original participant ID to be sanitized.
    sanitized_session_id : str
        The sanitized session ID label.
    sanitized_participant_id : str
        The sanitized participant ID label.
    """

    sanitization_config: SanitizationConfig
    sanitization_file_path: pathlib.Path
    original_participant_id: str
    original_session_id: str

    def model_post_init(self, context: typing.Any, /) -> None:
        self.sanitization_file_path.parent.mkdir(parents=True, exist_ok=True)
        with self.sanitization_file_path.open(mode="w") as file_stream:
            file_stream.write(self.model_dump_json())

    @staticmethod
    def _sanitize_label(label: str) -> str:
        """Sanitize a generic entity label by replacing non-alphanumeric characters with plus signs."""
        sanitized_label = re.sub(pattern=r"[^a-zA-Z0-9]", repl="+", string=label).removeprefix("+").removesuffix("+")
        return sanitized_label

    @pydantic.computed_field
    @property
    def sanitized_participant_id(self) -> str | None:
        if self.sanitization_config.sub_labels is False:
            return self.original_participant_id
        return self._sanitize_label(label=self.original_participant_id)

    @pydantic.computed_field
    @property
    def sanitized_session_id(self) -> str | None:
        if self.sanitization_config.ses_labels is False:
            return self.original_session_id
        return self._sanitize_label(label=self.original_session_id)



import pathlib

import nwb2bids


def test_create_file_tree(testing_files_directory: pathlib.Path) -> None:
    test_directory = testing_files_directory / "test_file_tree_from_dict"
    test_directory.mkdir(exist_ok=True)

    test_structure = {
        "a": {
            "b": {
                "file1.txt": "Content of file 1",
                "file2.txt": "Content of file 2",
            },
            "c": {
                "d": {
                    "file3.txt": "Content of file 3",
                },
            },
        },
        "file4.txt": "Content of file 4",
    }
    nwb2bids.testing.create_file_tree(directory=test_directory, structure=test_structure)

    expected_structure = {
        test_directory: {
            "directories": {"a"},
            "files": {"file4.txt"},
        },
        test_directory
        / "a": {
            "directories": {"b", "c"},
            "files": set(),
        },
        test_directory
        / "a"
        / "b": {
            "directories": set(),
            "files": {"file1.txt", "file2.txt"},
        },
        test_directory
        / "a"
        / "c": {
            "directories": {"d"},
            "files": set(),
        },
        test_directory
        / "a"
        / "c"
        / "d": {
            "directories": set(),
            "files": {"file3.txt"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(directory=test_directory, expected_structure=expected_structure)



"""Integration tests for the primary `convert_nwb_dataset` function."""

import json
import pathlib

import pytest

import nwb2bids


def test_minimal_convert_nwb_dataset_from_directory(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path.parent]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    assert dataset_converter.run_config.notifications_json_file_path.exists()
    with dataset_converter.run_config.notifications_json_file_path.open(mode="r") as file_stream:
        notifications_json = json.load(fp=file_stream)
    expected_notification_json = []
    assert notifications_json == expected_notification_json


def test_minimal_convert_nwb_dataset_from_file_path(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_ecephys_tutorial_convert_nwb_dataset(
    ecephys_tutorial_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [ecephys_tutorial_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-001"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-001": {
            "directories": {"ses-A"},
            "files": {"sub-001_sessions.json", "sub-001_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-001"
        / "ses-A": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-001"
        / "ses-A"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-001_ses-A_ecephys.nwb",
                "sub-001_ses-A_channels.tsv",
                "sub-001_ses-A_channels.json",
                "sub-001_ses-A_electrodes.tsv",
                "sub-001_ses-A_electrodes.json",
                "sub-001_ses-A_probes.tsv",
                "sub-001_ses-A_probes.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    probes_tsv_file_path = temporary_bids_directory / "sub-001" / "ses-A" / "ecephys" / "sub-001_ses-A_probes.tsv"
    probes_tsv_lines = probes_tsv_file_path.read_text().splitlines()
    expected_probe_tsv_lines = [
        "probe_name\ttype\tmanufacturer\tdescription",
        (
            "ExampleProbe\tn/a\t`nwb2bids.testing` module\tThis is an example ecephys probe "
            "used for demonstration purposes."
        ),
    ]
    assert probes_tsv_lines == expected_probe_tsv_lines

    electrodes_tsv_file_path = (
        temporary_bids_directory / "sub-001" / "ses-A" / "ecephys" / "sub-001_ses-A_electrodes.tsv"
    )
    electrodes_tsv_lines = electrodes_tsv_file_path.read_text().splitlines()
    expected_electrodes_tsv_lines = [
        "name\tprobe_name\themisphere\tx\ty\tz\timpedance\tshank_id\tlocation",
        "e000\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e001\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e002\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e003\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e004\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e005\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e006\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
        "e007\tExampleProbe\tN/A\tN/A\tN/A\tN/A\t150.0\tExampleShank\thippocampus",
    ]
    assert electrodes_tsv_lines == expected_electrodes_tsv_lines

    channels_tsv_file_path = temporary_bids_directory / "sub-001" / "ses-A" / "ecephys" / "sub-001_ses-A_channels.tsv"
    channels_tsv_lines = channels_tsv_file_path.read_text().splitlines()
    expected_channels_tsv_lines = [
        "channel_name\treference\ttype\tunit\tsampling_frequency\tstream_id\tgain",
        "ch0\te0\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch1\te1\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch2\te2\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch3\te3\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch4\te4\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch5\te5\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch6\te6\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
        "ch7\te7\tN/A\tV\t30000.0\tExampleElectricalSeries\t3.02734375e-06",
    ]
    assert channels_tsv_lines == expected_channels_tsv_lines


def test_ecephys_minimal_convert_nwb_dataset(
    ecephys_minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [ecephys_minimal_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-001"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-001": {
            "directories": {"ses-A"},
            "files": {"sub-001_sessions.json", "sub-001_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-001"
        / "ses-A": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-001"
        / "ses-A"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-001_ses-A_ecephys.nwb",
                "sub-001_ses-A_channels.tsv",
                "sub-001_ses-A_channels.json",
                "sub-001_ses-A_electrodes.tsv",
                "sub-001_ses-A_electrodes.json",
                "sub-001_ses-A_probes.tsv",
                "sub-001_ses-A_probes.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    probes_tsv_file_path = temporary_bids_directory / "sub-001" / "ses-A" / "ecephys" / "sub-001_ses-A_probes.tsv"
    probes_tsv_lines = probes_tsv_file_path.read_text().splitlines()
    expected_probe_tsv_lines = [
        "probe_name\ttype\tdescription",
        "ExampleProbe\tn/a\tThis is an example probe used for demonstration purposes.",
    ]
    assert probes_tsv_lines == expected_probe_tsv_lines

    electrodes_tsv_file_path = (
        temporary_bids_directory / "sub-001" / "ses-A" / "ecephys" / "sub-001_ses-A_electrodes.tsv"
    )
    electrodes_tsv_lines = electrodes_tsv_file_path.read_text().splitlines()
    expected_electrodes_tsv_lines = [
        "name\tprobe_name\themisphere\tx\ty\tz\timpedance\tshank_id\tlocation",
        "e000\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e001\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e002\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e003\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e004\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e005\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e006\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
        "e007\tExampleProbe\tN/A\tN/A\tN/A\tN/A\tN/A\tExampleShank\tunknown",
    ]
    assert electrodes_tsv_lines == expected_electrodes_tsv_lines

    channels_tsv_file_path = temporary_bids_directory / "sub-001" / "ses-A" / "ecephys" / "sub-001_ses-A_channels.tsv"
    channels_tsv_lines = channels_tsv_file_path.read_text().splitlines()
    expected_channels_tsv_lines = [
        "channel_name\treference\ttype\tunit",
        "ch0\te0\tN/A\tV",
        "ch1\te1\tN/A\tV",
        "ch2\te2\tN/A\tV",
        "ch3\te3\tN/A\tV",
        "ch4\te4\tN/A\tV",
        "ch5\te5\tN/A\tV",
        "ch6\te6\tN/A\tV",
        "ch7\te7\tN/A\tV",
    ]
    assert channels_tsv_lines == expected_channels_tsv_lines


# TODO: in follow-up, add test that checks if manufacturer column is dropped when empty


def test_implicit_bids_directory(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path, monkeypatch: pytest.MonkeyPatch
):
    implicit_bids_directory = temporary_bids_directory / "test_convert_nwb_dataset_implicit_bids"
    monkeypatch.chdir(temporary_bids_directory)

    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        implicit_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        implicit_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        implicit_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        implicit_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=implicit_bids_directory, expected_structure=expected_structure
    )


def test_implicit_bids_directory_with_relative_nwb_paths(
    minimal_nwbfile_path: pathlib.Path, temporary_run_directory: pathlib.Path, monkeypatch: pytest.MonkeyPatch
):
    """Test implicit bids_directory (cwd) with relative nwb_paths."""
    monkeypatch.chdir(temporary_run_directory)

    nwb_relative = minimal_nwbfile_path.relative_to(temporary_run_directory.parent)
    nwb_paths = [pathlib.Path("..") / nwb_relative]

    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_run_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_run_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_run_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_run_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_ecephys.nwb"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_run_directory, expected_structure=expected_structure
    )



"""
Collection of sanitization procedures to apply per field.

'Sanitization' in this sense means taking the NWB values for various fields, such as labels, and transforming them
to conform to BIDS.

Enabling more types of sanitization increases the validity of the resulting BIDS dataset.
"""

from ._configuration import SanitizationConfig
from ._sanitization import Sanitization

__all__ = [
    "Sanitization",
    "SanitizationConfig",
]



import pathlib
import typing

import numpy
import pydantic
import pynwb
import pynwb.testing.mock.ecephys
import pynwb.testing.mock.file
import pynwb.testing.mock.icephys


def get_tutorial_directory() -> pathlib.Path:
    tutorial_dir = pathlib.Path.home() / "nwb2bids_tutorials"
    tutorial_dir.mkdir(exist_ok=True)
    return tutorial_dir


def _generate_ecephys_file(*, nwbfile_path: pathlib.Path, subject_id: str = "001", session_id: str = "A") -> None:
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(
        session_id=session_id,
        session_description=(
            "An example NWB file containing ecephys neurodata types - for use in the nwb2bids tutorials."
        ),
    )

    subject = pynwb.file.Subject(
        subject_id=subject_id,
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    probe = pynwb.testing.mock.ecephys.mock_Device(
        name="ExampleProbe",
        description="This is an example ecephys probe used for demonstration purposes.",
        manufacturer="`nwb2bids.testing` module",
        nwbfile=nwbfile,
    )
    shank = pynwb.testing.mock.ecephys.mock_ElectrodeGroup(
        name="ExampleShank",
        description="This is an example electrode group (shank) used for demonstration purposes.",
        location="hippocampus",
        device=probe,
        nwbfile=nwbfile,
    )

    number_of_electrodes = 8
    for index in range(number_of_electrodes):
        nwbfile.add_electrode(imp=150_000.0, location="hippocampus", group=shank, filtering="HighpassFilter")
    electrodes = nwbfile.create_electrode_table_region(
        region=list(range(number_of_electrodes)),
        description="A `DynamicTableRegion` referring to all electrodes in this file.",
    )

    pynwb.testing.mock.ecephys.mock_ElectricalSeries(
        name="ExampleElectricalSeries",
        description=(
            "An example electrical series that represents data which could have been "
            "read off of the channels of an ecephys probe."
        ),
        data=numpy.ones(shape=(10, number_of_electrodes)),
        electrodes=electrodes,
        rate=30_000.0,
        nwbfile=nwbfile,
        conversion=0.00000302734375,
    )

    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)


def _generate_icephys_file(*, nwbfile_path: pathlib.Path, subject_id: str = "001", session_id: str = "A") -> None:
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(
        session_id=session_id,
        session_description=(
            "An example NWB file containing icephys neurodata types - for use in the nwb2bids tutorials."
        ),
    )

    subject = pynwb.file.Subject(
        subject_id=subject_id,
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    # Icephys 'probes'
    probe1 = pynwb.testing.mock.icephys.mock_Device(
        name="pipette01",
        description="This is an example icephys probe used for demonstration purposes.",
        manufacturer="Sutter",
        nwbfile=nwbfile,
    )
    probe2 = pynwb.testing.mock.icephys.mock_Device(
        name="pipette02",
        description="This is an example icephys probe used for demonstration purposes.",
        manufacturer="Sutter",
        nwbfile=nwbfile,
    )
    probe3 = pynwb.testing.mock.icephys.mock_Device(
        name="pipette03",
        description="This is an example icephys probe used for demonstration purposes.",
        manufacturer="WPI",
        nwbfile=nwbfile,
    )

    # Icephys electrodes
    electrode1 = nwbfile.create_icephys_electrode(
        name="patch01",
        description="This is an example icephys electrode used for demonstration purposes.",
        device=probe1,
    )
    electrode2 = nwbfile.create_icephys_electrode(
        name="patch02",
        description="This is an example icephys electrode used for demonstration purposes.",
        device=probe2,
    )
    electrode3 = nwbfile.create_icephys_electrode(
        name="sharp01",
        description="This is an example icephys electrode used for demonstration purposes.",
        device=probe3,
    )

    # Icephys series
    series1 = pynwb.testing.mock.icephys.mock_CurrentClampSeries(
        name="ExampleCurrentClampSeries1",
        data=[-70, -60, -50, -40, -30],
        conversion=1e-3,
        resolution=numpy.nan,
        rate=20e3,
        electrode=electrode1,
        gain=0.01,
        bias_current=1e-12,
        bridge_balance=70e6,
        capacitance_compensation=1e-12,
    )
    series2 = pynwb.testing.mock.icephys.mock_CurrentClampSeries(
        name="ExampleCurrentClampSeries2",
        data=[-10, 0, 10, 20, 30],
        conversion=1e-3,
        resolution=numpy.nan,
        rate=20e3,
        electrode=electrode2,
        gain=0.01,
        bias_current=1e-12,
        bridge_balance=70e6,
        capacitance_compensation=1e-12,
    )
    nwbfile.add_intracellular_recording(electrode=electrode1, response=series1)
    nwbfile.add_intracellular_recording(electrode=electrode2, response=series2)

    pynwb.testing.mock.icephys.mock_VoltageClampSeries(
        name="ExampleVoltageClampSeries1",
        data=[1.0, 2.0, 3.0, 4.0, 5.0],
        conversion=1e-12,
        resolution=numpy.nan,
        rate=20e3,
        electrode=electrode3,
        gain=5e-12,
        capacitance_slow=100e-12,
        resistance_comp_correction=70.0,
        nwbfile=nwbfile,
    )

    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)


@pydantic.validate_call
def generate_ephys_tutorial(
    *,
    mode=typing.Literal["file", "dataset"],
    output_directory: pydantic.DirectoryPath | None = None,
    modality: typing.Literal["ecephys", "icephys"] = "ecephys",
) -> pathlib.Path:
    if output_directory is None:
        tutorial_dir = get_tutorial_directory()
        output_directory = tutorial_dir / f"{modality}_tutorial_{mode}"
        output_directory.mkdir(exist_ok=True)

    if mode == "file":
        nwbfile_path = output_directory / f"{modality}.nwb"

        if modality == "ecephys":
            _generate_ecephys_file(nwbfile_path=nwbfile_path)
        else:
            _generate_icephys_file(nwbfile_path=nwbfile_path)

        return nwbfile_path

    subdir = output_directory / "some_sessions"
    subdir.mkdir(exist_ok=True)
    index_to_paths = {
        0: subdir / f"{modality}_session_1.nwb",
        1: subdir / f"{modality}_session_2.nwb",
        2: output_directory / f"{modality}_session_3.nwb",
        3: output_directory / "DO_NOT_CONVERT.nwb",
    }
    index_to_subject_id = {
        0: "001",
        1: "001",
        2: "002",
        3: "003",
    }
    index_to_session_id = {
        0: "A",
        1: "B",
        2: "C",
        3: "D",
    }

    for index in range(4):
        nwbfile_path = index_to_paths[index]

        if modality == "ecephys":
            _generate_ecephys_file(
                nwbfile_path=nwbfile_path, subject_id=index_to_subject_id[index], session_id=index_to_session_id[index]
            )
        else:
            _generate_icephys_file(
                nwbfile_path=nwbfile_path, subject_id=index_to_subject_id[index], session_id=index_to_session_id[index]
            )

    return output_directory



"""
Integration tests for edge cases of the primary `convert_nwb_dataset` function.

Namely, the cases of:
  - A dataset with additional metadata and the case of a dataset.
  - A dataset without a session ID.
"""

import os
import pathlib

import pandas
import pytest

import nwb2bids


def test_convert_nwb_dataset_with_additional_metadata(
    minimal_nwbfile_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
    additional_metadata_file_path: pathlib.Path,
):
    nwb_paths = [minimal_nwbfile_path]
    run_config = nwb2bids.RunConfig(
        bids_directory=temporary_bids_directory, additional_metadata_file_path=additional_metadata_file_path
    )
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"participants.json", "participants.tsv", "dataset_description.json"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_ecephys.nwb"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_convert_nwb_dataset_on_mock_datalad_dataset(
    mock_datalad_dataset: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [mock_datalad_dataset]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"participants.json", "participants.tsv", "dataset_description.json"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_ecephys.nwb"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_convert_nwb_dataset_on_mock_datalad_dataset_with_broken_symlink(
    mock_datalad_dataset: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    broken_symlink = mock_datalad_dataset / "broken_symlink.nwb"
    broken_symlink.symlink_to(target="non_existent_file.nwb")

    nwb_paths = [mock_datalad_dataset]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"participants.json", "participants.tsv", "dataset_description.json"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_ecephys.nwb"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_convert_nwb_dataset_with_subject_mismatch(
    minimal_nwbfile_path: pathlib.Path,
    minimal_mismatch_nwbfile_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
):
    nwb_paths = [minimal_nwbfile_path, minimal_mismatch_nwbfile_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory)
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert len(dataset_converter.notifications) == 1

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"participants.json", "participants.tsv", "dataset_description.json"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456", "ses-4567"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_ecephys.nwb"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-4567": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-4567"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-4567_ecephys.nwb"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    participants_tsv_file_path = temporary_bids_directory / "participants.tsv"
    actual_dataframe = pandas.read_csv(filepath_or_buffer=participants_tsv_file_path, sep="\t")
    expected_dataframe = pandas.DataFrame(
        {"participant_id": {0: "sub-123"}, "species": {0: "Mus musculus, mouse"}, "sex": {0: "M"}}
    )
    pandas.testing.assert_frame_equal(left=actual_dataframe, right=expected_dataframe)


def test_symlink_resolves_correctly_with_relative_path(
    minimal_nwbfile_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test that symlinks created from relative paths resolve to the correct target."""
    monkeypatch.chdir(minimal_nwbfile_path.parent)
    relative_nwb_path = pathlib.Path(minimal_nwbfile_path.name)

    nwb_paths = [relative_nwb_path]
    run_config = nwb2bids.RunConfig(bids_directory=temporary_bids_directory, file_mode="symlink")
    dataset_converter = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, run_config=run_config)
    assert not any(dataset_converter.notifications)

    symlink_path = temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_ecephys.nwb"

    assert symlink_path.is_symlink(), "Expected a symlink to be created"
    assert symlink_path.resolve() == minimal_nwbfile_path.resolve(), "Symlink does not resolve to original file"
    assert os.path.exists(symlink_path), "Symlink is broken (target does not exist)"



"""
Mostly for demonstration purposes.

The regular expression for labels is trusted to work over all basic cases.
Any edge cases may also be added here to ensure confidence of coverage.
"""

import pathlib

import pytest

import nwb2bids


@pytest.mark.parametrize(
    "participant_id, sanitization_config, expected",
    [
        ("Mouse 12", nwb2bids.sanitization.SanitizationConfig(), "Mouse 12"),
        ("Mouse 12", nwb2bids.sanitization.SanitizationConfig(sub_labels=True), "Mouse+12"),
        (
            "Raw/data?sub-01@today",
            nwb2bids.sanitization.SanitizationConfig(sub_labels=True),
            "Raw+data+sub+01+today",
        ),
    ],
)
def test_sanitize_participant_id(participant_id, sanitization_config, expected, temporary_run_directory: pathlib.Path):
    sanitization_file_path = temporary_run_directory / "test_sanitize_session_id_results.txt"
    sanitization = nwb2bids.sanitization.Sanitization(
        sanitization_config=sanitization_config,
        sanitization_file_path=sanitization_file_path,
        original_participant_id=participant_id,
        original_session_id="12_02_2025",
    )
    assert sanitization.sanitized_participant_id == expected


@pytest.mark.parametrize(
    "session_id, sanitization_config, expected",
    [
        (
            "Session 12 subject 5",
            nwb2bids.sanitization.SanitizationConfig(),
            "Session 12 subject 5",
        ),
        # TODO: think about if subject info in session IDs deserves special cleaning
        (
            "Session 12 subject 5",
            nwb2bids.sanitization.SanitizationConfig(ses_labels=True),
            "Session+12+subject+5",
        ),
        # TODO: think about if timestamps deserve special cleaning
        (
            "2025-02-03T01:08:12.1236",
            nwb2bids.sanitization.SanitizationConfig(ses_labels=True),
            "2025+02+03T01+08+12+1236",
        ),
        (
            "12_02_2025",
            nwb2bids.sanitization.SanitizationConfig(ses_labels=True),
            "12+02+2025",
        ),
    ],
)
def test_sanitize_session_id(session_id, sanitization_config, expected, temporary_run_directory: pathlib.Path):
    sanitization_file_path = temporary_run_directory / "test_sanitize_session_id_results.txt"
    sanitization = nwb2bids.sanitization.Sanitization(
        sanitization_config=sanitization_config,
        sanitization_file_path=sanitization_file_path,
        original_participant_id="0",
        original_session_id=session_id,
    )
    assert sanitization.sanitized_session_id == expected



import abc
import typing

import pydantic

from ._run_config import RunConfig
from ..notifications import Notification


class BaseConverter(pydantic.BaseModel, abc.ABC):
    run_config: typing.Annotated[
        RunConfig, pydantic.Field(description="The configuration for this conversion run.", default_factory=RunConfig)
    ]
    _internal_notifications: list[Notification] = pydantic.PrivateAttr(default_factory=list)

    @abc.abstractmethod
    def extract_metadata(self) -> None:
        """
        Extract essential metadata used by the BIDS standard from the source NWB files.
        """
        message = f"The `extract_metadata` method has not been implemented by the `{self.__class__.__name__}` class."
        raise NotImplementedError(message)



import json
import pathlib
import typing
from typing import Any

import pandas
import pydantic
import pynwb
import typing_extensions

from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel
from ..notifications import Category, DataStandard, Notification, Severity


class Probe(BaseMetadataModel):
    probe_name: str
    type: str = "n/a"
    manufacturer: str | None = None


class ProbeTable(BaseMetadataContainerModel):
    probes: list[Probe]
    modality: typing.Literal["ecephys", "icephys"]

    def _check_fields(self) -> None:
        # Check if values are specified
        self._internal_notifications = []

        probes_missing_description = [probe for probe in self.probes if probe.description is None]
        for probe_missing_description in probes_missing_description:
            self._internal_notifications.append(
                Notification(
                    title="Missing description",
                    reason="A basic description of this field is recommended to improve contextual understanding.",
                    solution="Add a description to the field.",
                    field=f"nwbfile.devices.{probe_missing_description.probe_name}",
                    source_file_paths=[],  # TODO: figure out better way of handling
                    data_standards=[DataStandard.BIDS, DataStandard.NWB],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.INFO,
                )
            )

    def model_post_init(self, context: Any, /) -> None:
        self._check_fields()

    @pydantic.computed_field
    @property
    def notifications(self) -> list[Notification]:
        """
        All notifications from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        notifications = [notification for probe in self.probes for notification in probe.notifications]
        notifications += self._internal_notifications
        notifications.sort(
            key=lambda notification: (-notification.category.value, -notification.severity.value, notification.title)
        )
        return notifications

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        has_ecephys_probes = nwbfile.electrodes is not None
        has_icephys_probes = any(nwbfile.icephys_electrodes)
        if not has_ecephys_probes and not has_icephys_probes:
            return None

        if has_ecephys_probes and has_icephys_probes:
            message = (
                "Converting probe metadata when there are both ecephys and icephys types has not yet been implemented."
            )
            raise NotImplementedError(message)

        modality = "ecephys" if has_ecephys_probes else "icephys"
        if modality == "ecephys":
            electrode_groups = nwbfile.electrodes["group"][:]
            unique_devices = {electrode_group.device for electrode_group in electrode_groups}
        else:
            icephys_electrodes = nwbfile.icephys_electrodes.values()
            unique_devices = {electrode.device for electrode in icephys_electrodes}

        probes = [
            Probe(
                probe_name=device.name,
                type="n/a",  # TODO
                manufacturer=device.manufacturer,
                description=device.description,
                # TODO: handle more extra custom columns
            )
            for device in unique_devices
        ]
        return cls(probes=probes, modality=modality)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path) -> None:
        """
        Save the probes information to a TSV file.

        Parameters
        ----------
        file_path : path
            The path to the output TSV file.
        """
        data = []
        for probe in self.probes:
            model_dump = probe.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)
        data_frame = data_frame.dropna(axis=1, how="all")
        data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the probes information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        file_path = pathlib.Path(file_path)

        with file_path.open(mode="w") as file_stream:
            json.dump(
                obj=dict(),  # TODO
                fp=file_stream,
                indent=4,
            )



import pathlib


def _get_nwb2bids_home_directory() -> pathlib.Path:
    """
    Get the home directory used by the `nwb2bids` project.

    Only light-weight configuration files are kept here.
    Heavier contents (such as text files, logs, etc.) are stored in the cache directory.
    """
    user_home = pathlib.Path.home()
    nwb2bids_home = user_home / ".nwb2bids"
    nwb2bids_home.mkdir(exist_ok=True)
    return nwb2bids_home



import json
import pathlib
import typing

import numpy
import pandas
import pydantic
import pynwb
import typing_extensions

from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel
from ..notifications import Notification


class Electrode(BaseMetadataModel):
    name: str
    probe_name: str
    hemisphere: str = "N/A"
    x: float = numpy.nan
    y: float = numpy.nan
    z: float = numpy.nan
    impedance: float = numpy.nan  # in kOhms
    shank_id: str = "N/A"
    location: str | None = None

    def __eq__(self, other: typing_extensions.Self) -> bool:
        if not isinstance(other, Electrode):
            return False

        self_dump = self.model_dump()
        other_dump = other.model_dump()

        if self_dump.keys() != other_dump.keys():
            return False

        for key in self_dump:
            self_val = self_dump[key]
            other_val = other_dump[key]

            # Handle NaN comparison (NaN == NaN should be true)
            if isinstance(self_val, float) and isinstance(other_val, float):
                if numpy.isnan(self_val) and numpy.isnan(other_val):
                    continue

            if self_val != other_val:
                return False

        return True


class ElectrodeTable(BaseMetadataContainerModel):
    electrodes: list[Electrode]
    modality: typing.Literal["ecephys", "icephys"]

    @pydantic.computed_field
    @property
    def notifications(self) -> list[Notification]:
        """
        All notifications from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        notifications = [notification for electrode in self.electrodes for notification in electrode.notifications]
        notifications.sort(
            key=lambda notification: (-notification.category.value, -notification.severity.value, notification.title)
        )
        return notifications

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        has_ecephys_electrodes = nwbfile.electrodes is not None
        has_icephys_electrodes = any(nwbfile.icephys_electrodes)
        if not has_ecephys_electrodes and not has_icephys_electrodes:
            return None

        if has_ecephys_electrodes and has_icephys_electrodes:
            message = (
                "Converting electrode metadata when there are both ecephys and icephys types "
                "has not yet been implemented."
            )
            raise NotImplementedError(message)

        modality = "ecephys" if has_ecephys_electrodes else "icephys"
        if modality == "ecephys":
            electrodes = [
                Electrode(
                    # Leading 'e' and padding are by convention from BEP32 examples
                    name=f"e{str(electrode.index[0]).zfill(3)}",
                    probe_name=electrode.group.iloc[0].device.name,
                    # TODO: hemisphere determination from additional metadata or possible lookup from a location map
                    x=electrode.x.iloc[0] if "x" in electrode else numpy.nan,
                    y=electrode.y.iloc[0] if "y" in electrode else numpy.nan,
                    z=electrode.z.iloc[0] if "z" in electrode else numpy.nan,
                    # Impedance must be in kOhms for BEP32 but NWB specifies Ohms
                    impedance=(
                        eimp_in_ohms / 1e3
                        if "imp" in electrode and not pandas.isna(eimp_in_ohms := electrode.imp.iloc[0])
                        else numpy.nan
                    ),
                    shank_id=electrode.group.iloc[0].name,
                    # TODO: pretty much only through additional metadata
                    # size=
                    # electrode_shape=
                    # material=
                    location=str(electrode.location.values[0]),
                    # TODO: add extra columns
                )
                for electrode in nwbfile.electrodes
            ]
        else:
            electrodes = [
                Electrode(
                    name=electrode.name,
                    probe_name=electrode.device.name,
                    # TODO: hemisphere, location, impedance may all have to be specified through custom columns
                    x=getattr(electrode, "x", numpy.nan),
                    y=getattr(electrode, "y", numpy.nan),
                    z=getattr(electrode, "z", numpy.nan),
                    # impedance=  # Impedance must be in kOhms for BEP32 but NWB specifies Ohms
                    # shank_id=
                    # TODO: pretty much only through additional metadata
                    # size=
                    # electrode_shape=
                    # material=
                    # location=
                    # TODO: some icephys specific ones (would NOT use the ecephys electrode table anyway...)
                    # Probably better off in a designated model
                    # pipette_solution=
                    # internal_pipette_diameter=
                    # external_pipette_diameter=
                    # TODO: add extra columns
                )
                for electrode in nwbfile.icephys_electrodes.values()
            ]
        return cls(electrodes=electrodes, modality=modality)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path) -> None:
        """
        Write the electrode data to a TSV file.

        Parameters
        ----------
        file_path : path
            The path to the output TSV file.
        """
        data = []
        for electrode in self.electrodes:
            model_dump = electrode.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)

        # Many columns are 'required' by BEP32 but are not always present in the source files or known at all
        required_columns = ["x", "y", "z", "impedance"]
        for column in required_columns:
            data_frame[column] = data_frame[column].fillna(value="N/A")
        data_frame = data_frame.dropna(axis=1, how="all")
        data_frame.to_csv(file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the electrode information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(
                obj=dict(),  # TODO
                fp=file_stream,
                indent=4,
            )
